#include "Selection.h"
#include <AppKit/NSAffineTransform.h>
#include <AppKit/NSEvent.h>
#include <AppKit/NSColor.h>
#include <AppKit/NSImage.h>
#include <AppKit/NSApplication.h>
#include <AppKit/NSCursor.h>
#include "PHGeometry.h"
#include "PHAffineTransform.h"
#include "AppController.h"
#include "SelectionCtrl.h"
#include "MouseTracker.h"
#include "CursorRectInfo.h"
#include <math.h>

#define INTEGRAL_SIZE

typedef enum { PHOutsideRect = 0, PHInsideRect, PHCornerRect, PHEdgeRect }
PHPHRectRegion;

const static int BW = 2; // boundary width / 2

@interface Selection (PrivateMethods)

- (void) stretchToFitMinimalSpan;
- (MouseTracker *) createMouseTrackerWithStartPoint: (NSPoint) pt;
- (int) areaOfRect: (NSRect) rc fromPoint: (NSPoint) pt;

@end // Selection (PrivateMethods)


@implementation Selection

- (id) init
{
    _fragment = NSZeroRect;
    _sketch = NSZeroRect;
    _selection = NSZeroRect;
    _previousSelection = NSZeroRect;

    _is_changing = NO;

    _xform = nil;

    _tracker = nil;

    return self;
}

- (void) dealloc
{
    RELEASE(_xform);
    RELEASE(_tracker);
}

- (void) clear
{
    RELEASE(_xform);
    _xform = nil;
    RELEASE(_tracker);
    _tracker = nil;

    _selection = NSZeroRect;
    _previousSelection = _sketch; // pretend that the whole area changed
    _fragment = NSZeroRect;
    _sketch = NSZeroRect;
}

- (void) setFragment: (NSRect) fragment xform: (NSAffineTransform *) xform
{
    if (_xform != xform)
    {
        BOOL curr_is_empty = (_xform == nil);
        
        _fragment = fragment;

        NSRect orig_selection;
        if (curr_is_empty == NO)
        {
            NSAffineTransform * reverse = PHReverseTransform(_xform);
            orig_selection = [reverse transformRect: _selection];
        }
        
        RELEASE(_xform);
        _xform = xform;
        RETAIN(_xform);

        // need to convert _start, _end and _sketch
        if (curr_is_empty == NO)
        {
            _selection = [_xform transformRect: orig_selection];
        }
        _sketch = [_xform transformRect: _fragment];
        
#ifdef INTEGRAL_SIZE
        _sketch = NSIntegralRect(_sketch);
        _sketch.origin = NSZeroPoint;
        _selection = NSIntegralRect(_selection);
#endif //  INTEGRAL_SIZE

    }
}

- (void) mouseDown: (NSPoint) pt
{
    _is_changing = YES;

    ASSIGN(_tracker, [self createMouseTrackerWithStartPoint: pt]);

    _selection = [_tracker selectionRect];

    // force full redraw if selection is invalid
    _previousSelection = [self valid] ? _selection : _sketch;
}

- (void) mouseDragged: (NSPoint) pt
{
    NSAssert(_tracker, @"_tracker == nil");

    _previousSelection = [_tracker selectionRect];
    _selection = [_tracker mouseDragged: pt];
}

- (void) mouseUp: (NSPoint) pt
{
    // I have to comment it because of low-pass filter: it adds "inertia"
    //[self mouseDragged: pt];

    // TODO: do that only in "predefined extent" mode.
    if ([self valid]) // if nonempty
        [self stretchToFitMinimalSpan];

    if (! [self valid])
        _previousSelection = _sketch; // force full redraw

    _is_changing = NO;
}

- (MouseTracker *) createMouseTrackerWithStartPoint: (NSPoint) pt
{
    MouseTracker * tracker = nil;

    int area = [self areaOfRect: _selection fromPoint: pt];

    Class cls;
    switch (area)
    {
        case PHInsideRect:
            cls = [MouseTrackerMoveAll class];
            break;
        case PHEdgeRect:
            cls = [MouseTrackerMoveCorner class];
            break;
        case PHOutsideRect:
            cls = [MouseTrackerNewSel class];
            break;
        default:
            NSAssert1(0, @"unknown rectangle area=%d", area);
            cls = [MouseTrackerNewSel class];
            break;
    }
    
    tracker = [[cls alloc] initWithFrame: _sketch selRect: _selection
                           start: pt];

    return AUTORELEASE(tracker);
}

- (int) areaOfRect: (NSRect) rc fromPoint: (NSPoint) pt
{
    
    NSRect left = NSMakeRect(NSMinX(rc) - BW, NSMinY(rc) - BW,
                             2*BW, 2*BW + rc.size.height);
    NSRect right = NSMakeRect(NSMaxX(rc) - BW, NSMinY(rc) - BW,
                              2*BW, 2*BW + rc.size.height);
    NSRect bottom = NSMakeRect(NSMinX(rc) - BW, NSMinY(rc) - BW,
                               2*BW + rc.size.width, 2*BW);
    NSRect top = NSMakeRect(NSMinX(rc) - BW, NSMaxY(rc) - BW,
                            2*BW + rc.size.width, 2*BW);

    BOOL on_edge =
        NSPointInRect(pt, left) || NSPointInRect(pt, right) ||
        NSPointInRect(pt, bottom) || NSPointInRect(pt, top);

    if (on_edge)
        return PHEdgeRect;
    else if (NSPointInRect(pt, rc))
        return PHInsideRect;
    else
        return PHOutsideRect;
        
}

- (NSArray *) cursorRects
{
    NSMutableArray * result = [NSMutableArray arrayWithCapacity: 2];
    NSRect rc;

    NSCursor * inner = [MouseTrackerMoveAll cursorType];
    NSCursor * edge = [MouseTrackerNewSel cursorType];
    
    rc = NSInsetRect(_selection, -BW, -BW);
    [result addObject: [CursorRectInfo infoWithRect: rc cursor: edge]];

    rc = NSInsetRect(_selection, BW, BW);
    [result addObject: [CursorRectInfo infoWithRect: rc cursor: inner]];

    return result;
}

- (BOOL) isOutOfSelPoint: (NSPoint) pt
{
    return
        ! [self valid] ||
        ! NSPointInRect(pt, NSInsetRect(_selection, -BW, -BW));
}

- (NSCursor *) trackingCursor
{
    return (_tracker) ? [_tracker cursor] : nil;
}

- (BOOL) valid
{
    return NSIsEmptyRect([self selectedInSketch]) == NO;
}

- (BOOL) isChanging
{
    return _is_changing;
}

- (BOOL) exists
{
    return _is_changing || [self valid];
}

- (void) stretchToFitMinimalSpan
{
    // If we do not allow stretching the fragment out,
    // the fragment's size must be bigger or equal to the predefined
    // scale-to size (the bound).
    // Here we enlarge the fragment if it's too small.
    
    float bound = [[[NSApp delegate] selectionCtrl] bound];
    
    if (bound > 0)
    {
        float m[6];
        [_xform getMatrix: m];
        double scale = sqrt(m[0] * m[0] + m[1] * m[1]);
        float min_span = bound * scale; // sketch coordinates

        float width = _selection.size.width;
        float height = _selection.size.height;
        float span = MAX(width, height); // sketch coordinates

        if (span < min_span)
        {
            // try to stretch selection to minimal span
            
            float shape = height/width;
            NSSize new_size; // sketch coordinates
            new_size.width = (shape < 1) ? min_span : min_span / shape;
            new_size.height = (shape < 1) ? min_span * shape : min_span;

            //NSLog(@"fragment=%@", NSStringFromRect(_fragment));
            //NSLog(@"sketch=%@", NSStringFromRect(_sketch));
            
            // adjust size only if the new size fits into sketch
            if (new_size.width <= _sketch.size.width &&
                new_size.height <= _sketch.size.height)
            {
                float x_mid = NSMidX(_selection);
                float y_mid = NSMidY(_selection);

                NSRect tmp = NSMakeRect(x_mid, y_mid, 0, 0);
                tmp = NSInsetRect(tmp, -new_size.width/2, -new_size.height/2);
                tmp = PHPushRectInsideBoundaries(tmp, _sketch);
                _selection = tmp;

#ifdef INTEGRAL_SIZE
                _selection = NSIntegralRect(_selection);
#endif // INTEGRAL_SIZE
            }
        }
    }
}

- (NSRect) selectedInSketch
{
    return NSIntersectionRect(_sketch, _selection);

    // _selection should never extend beyond _sketch, so can just
    // return _selection;
}

- (NSRect) updateInSketch
{
    return NSInsetRect(
        NSUnionRect(_previousSelection, [self selectedInSketch]),
        -1, -1);
}

- (NSRect) selectedInImage
{
    NSAffineTransform * reverse = PHReverseTransform(_xform);

    return NSIntersectionRect( _fragment,
                               [reverse transformRect: _selection] );
}

- (NSImage *) mask
{
    NSImage * mask = [[NSImage alloc] initWithSize: _sketch.size];
    
    NSColor * blur = [NSColor colorWithCalibratedWhite: 1.0 alpha: 0.57];
    NSColor * transp = [NSColor colorWithCalibratedWhite: 1.0 alpha: 0.0];

    [mask lockFocus];
    
    [blur set];

    NSRectFillUsingOperation(_sketch, NSCompositeCopy);

    [transp set];

    NSRect selectionRect = [self selectedInSketch];
    
    NSRectFillUsingOperation(selectionRect, NSCompositeCopy);
    
    [mask unlockFocus];

    return AUTORELEASE(mask);
}

- (NSImage *) maskForRect: (NSRect) rect
{
    NSImage * mask = [[NSImage alloc] initWithSize: rect.size];

    NSColor * blur = [NSColor colorWithCalibratedWhite: 1.0 alpha: 0.57];
    NSColor * transp = [NSColor colorWithCalibratedWhite: 0.0 alpha: 0.0];

    [mask lockFocus];
    
    [blur set];

    NSRect maskRect = NSMakeRect(0, 0, rect.size.width, rect.size.height);

    // sketch rect in mask coordinates
    NSRect imageRect = NSOffsetRect(_sketch, -rect.origin.x, -rect.origin.y);
    
    NSRectFillUsingOperation(NSIntersectionRect(maskRect, imageRect),
                             NSCompositeCopy);

    [transp set];

    // in sketch coordinates
    NSRect selectionRect = [self selectedInSketch];

    // convert to mask coordinates
    NSRect maskSelRect = NSOffsetRect(selectionRect,
                                      -rect.origin.x, -rect.origin.y);

    maskSelRect = NSIntersectionRect(maskSelRect, maskRect);
    
    NSRectFillUsingOperation(NSIntersectionRect(maskSelRect, imageRect),
                             NSCompositeCopy);
    
    [mask unlockFocus];

    return AUTORELEASE(mask);
}

@end // implementation Selection
